# test crs creation and conversion from one crs to another

import pytest
import numpy as np
import resqpy.model as rq
import resqpy.crs as rqc
import resqpy.olio.uuid as bu


def test_crs():

   # create some coordinate reference systems
   model = rq.Model(new_epc = True, create_basics = True)
   crs_default = rqc.Crs(model)
   assert crs_default.null_transform
   crs_m = rqc.Crs(model, xy_units = 'm', z_units = 'm')
   crs_ft = rqc.Crs(model, xy_units = 'ft', z_units = 'ft')
   crs_mixed = rqc.Crs(model, xy_units = 'm', z_units = 'ft')
   crs_offset = rqc.Crs(model, xy_units = 'm', z_units = 'm', x_offset = 100.0, y_offset = -100.0, z_offset = -50.0)
   assert not crs_offset.null_transform
   crs_elevation = rqc.Crs(model, z_inc_down = 'False')

   # check that distincitveness is recognised
   assert crs_default.is_equivalent(crs_m)
   assert not crs_m.is_equivalent(crs_ft)
   assert not crs_mixed.is_equivalent(crs_m)
   assert not crs_m.is_equivalent(crs_offset)
   assert not crs_m.is_equivalent(crs_elevation)

   # create some xml
   crs_m.create_xml()
   crs_offset.create_xml()

   # test conversion
   ft_to_m = 0.3048

   a = np.empty((10, 3))
   a[:, 0] = np.random.random(10) * 5.0e5
   a[:, 1] = np.random.random(10) * 10.0e5
   a[:, 2] = np.random.random(10) * 4.0e3

   b = a.copy()
   crs_m.convert_array_from(crs_default, a)
   assert np.max(np.abs(b - a)) < 1.0e-6
   a[:] = b
   crs_m.convert_array_to(crs_m, a)
   assert np.all(a == b)
   crs_ft.convert_array_from(crs_m, a)
   assert np.max(np.abs(b - a * ft_to_m)) < 1.0e-6
   crs_ft.convert_array_to(crs_m, a)
   assert np.max(np.abs(b - a)) < 1.0e-6
   a[:] = b
   crs_m.local_to_global_array(a)
   assert np.max(np.abs(b - a)) < 1.0e-6
   a[:] = b
   crs_offset.global_to_local_array(a)
   a[:, 0] += 100.0
   a[:, 1] -= 100.0
   a[:, 2] -= 50.0
   assert np.max(np.abs(b - a)) < 1.0e-6

   # todo: test rotation


def test_crs_reuse():
   model = rq.Model(new_epc = True, create_basics = True)
   crs_a = rqc.Crs(model)
   crs_a.create_xml()
   crs_b = rqc.Crs(model)
   crs_b.create_xml()
   assert len(model.parts(obj_type = 'LocalDepth3dCrs')) == 1
   assert crs_a == crs_b
   assert bu.matching_uuids(crs_a.uuid, crs_b.uuid)
   crs_c = rqc.Crs(model, z_inc_down = False)
   crs_c.create_xml()
   assert len(model.parts(obj_type = 'LocalDepth3dCrs')) == 2
   assert crs_c != crs_a
   assert not bu.matching_uuids(crs_c.uuid, crs_a.uuid)
   crs_d = rqc.Crs(model, z_units = 'ft')
   crs_d.create_xml()
   assert len(model.parts(obj_type = 'LocalDepth3dCrs')) == 3
   crs_e = rqc.Crs(model, z_inc_down = False)
   crs_e.create_xml()
   assert len(model.uuids(obj_type = 'LocalDepth3dCrs')) == 3
   assert crs_e == crs_c
   assert bu.matching_uuids(crs_e.uuid, crs_c.uuid)
   crs_f = rqc.Crs(model)
   crs_f.create_xml(reuse = False)
   assert len(model.parts(obj_type = 'LocalDepth3dCrs')) == 4
   assert crs_f == crs_a
   assert not bu.matching_uuids(crs_f.uuid, crs_a.uuid)

   # todo: test parent epsg code equivalence



import pytest

import resqpy.model as rq
import resqpy.lines
import resqpy.organize
import numpy as np
import os


def test_lines(example_model_and_crs):

   # Set up a Polyline
   title = 'Nazca'
   model, crs = example_model_and_crs
   line = resqpy.lines.Polyline(parent_model = model,
                                title = title,
                                set_crs = crs.uuid,
                                set_crsroot = crs.root,
                                set_bool = True,
                                set_coord = np.array([[0, 0, 0], [1, 1, 1]]))
   line.write_hdf5()
   line.create_xml()

   # Add a interpretation
   assert line.rep_int_root is None
   line.create_interpretation_and_feature(kind = 'fault')
   assert line.rep_int_root is not None

   # Check fault can be loaded in again
   model.store_epc()
   model = rq.Model(epc_file = model.epc_file)
   reload = resqpy.lines.Polyline(parent_model = model, uuid = line.uuid)
   assert reload.citation_title == title

   fault_interp = resqpy.organize.FaultInterpretation(model, uuid = line.rep_int_uuid)
   fault_feature = resqpy.organize.TectonicBoundaryFeature(model, uuid = fault_interp.feature_uuid)

   # Check title matches expected title
   assert fault_feature.feature_name == title


def test_lineset(example_model_and_crs):

   # Set up a PolylineSet
   title = 'Nazcas'
   model, crs = example_model_and_crs
   line1 = resqpy.lines.Polyline(parent_model = model,
                                 title = title,
                                 set_crs = crs.uuid,
                                 set_crsroot = crs.root,
                                 set_bool = True,
                                 set_coord = np.array([[0, 0, 0], [1, 1, 1]]))

   line2 = resqpy.lines.Polyline(parent_model = model,
                                 title = title,
                                 set_crs = crs.uuid,
                                 set_crsroot = crs.root,
                                 set_bool = True,
                                 set_coord = np.array([[0, 0, 0], [2, 2, 2]]))

   lines = resqpy.lines.PolylineSet(parent_model = model, title = title, polylines = [line1, line2])

   lines.write_hdf5()
   lines.create_xml()

   # Check lines can be loaded in again
   model.store_epc()
   model = rq.Model(epc_file = model.epc_file)
   reload = resqpy.lines.PolylineSet(parent_model = model, uuid = lines.uuid)
   assert len(reload.polys) == 2, f'Expected two polylines in the polylineset, found {len(reload.polys)}'
   assert (reload.count_perpol == [2, 2]).all(), f'Expected count per polyline to be [2,2], found {reload.count_perpol}'


def test_charisma(example_model_and_crs, test_data_path):
   # Set up a PolylineSet
   model, crs = example_model_and_crs
   charisma_file = test_data_path / "Charisma_example.txt"
   lines = resqpy.lines.PolylineSet(parent_model = model, charisma_file = str(charisma_file))
   lines.write_hdf5()
   lines.create_xml()

   model.store_epc()
   model = rq.Model(epc_file = model.epc_file)
   reload = resqpy.lines.PolylineSet(parent_model = model, uuid = lines.uuid)

   assert reload.title == 'Charisma_example'
   assert (reload.count_perpol == [
      4, 5, 4, 5, 5
   ]).all(), f"Expected count per polyline to be [4,5,4,5,5], found {reload.count_perpol}"
   assert len(reload.coordinates) == 23, f"Expected length of coordinates to be 23, found {len(reload.coordinates)}"


def test_irap(example_model_and_crs, test_data_path):
   # Set up a PolylineSet
   model, crs = example_model_and_crs
   irap_file = test_data_path / "IRAP_example.txt"
   lines = resqpy.lines.PolylineSet(parent_model = model, irap_file = str(irap_file))
   lines.write_hdf5()
   lines.create_xml()

   model.store_epc()
   model = rq.Model(epc_file = model.epc_file)
   reload = resqpy.lines.PolylineSet(parent_model = model, uuid = lines.uuid)

   assert reload.title == 'IRAP_example'
   assert (reload.count_perpol == [15]).all(), f"Expected count per polyline to be [15], found {reload.count_perpol}"
   assert len(reload.coordinates) == 15, f"Expected length of coordinates to be 15, found {len(reload.coordinates)}"



""" Shared fixtures for tests """

import logging

import pytest
from pathlib import Path
from shutil import copytree
import numpy as np
import pandas as pd
import os

from resqpy.model import Model, new_model
from resqpy.organize import WellboreFeature, WellboreInterpretation
from resqpy.well import Trajectory, MdDatum, WellboreFrame
from resqpy.crs import Crs
import resqpy.grid as grr
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp


@pytest.fixture(autouse = True)
def capture_logs(caplog):
   """Always capture log messages from respy"""

   caplog.set_level(logging.DEBUG, logger = "resqpy")


@pytest.fixture
def tmp_model(tmp_path):
   """Example resqpy model in a temporary directory unique to each test"""

   return new_model(str(tmp_path / 'tmp_model.epc'))


@pytest.fixture
def example_model_and_crs(tmp_model):
   """ Returns a fresh RESQML Model and Crs, in a temporary directory """

   # TODO: parameterise with m or feet
   xyz_uom = 'm'

   # Create a model with a coordinate reference system
   crs = Crs(parent_model = tmp_model, z_inc_down = True, xy_units = xyz_uom, z_units = xyz_uom)
   crs.create_xml()

   return tmp_model, crs


@pytest.fixture
def example_model_with_well(example_model_and_crs):
   """ Model with a single well with a vertical trajectory """

   wellname = 'well A'
   elevation = 100
   md_uom = 'm'

   model, crs = example_model_and_crs

   # Create a single well feature and interpretation
   well_feature = WellboreFeature(parent_model = model, feature_name = wellname)
   well_feature.create_xml()
   well_interp = WellboreInterpretation(parent_model = model, wellbore_feature = well_feature, is_drilled = True)
   well_interp.create_xml()

   # Create a measured depth datum
   datum = MdDatum(parent_model = model,
                   crs_uuid = crs.uuid,
                   location = (0, 0, -elevation),
                   md_reference = 'kelly bushing')
   datum.create_xml()

   # Create trajectory of a vertical well
   mds = np.array([0, 1000, 2000])
   zs = mds - elevation
   traj = Trajectory(parent_model = model,
                     md_datum = datum,
                     data_frame = pd.DataFrame(dict(MD = mds, X = 0, Y = 0, Z = zs)),
                     length_uom = md_uom,
                     represented_interp = well_interp)
   traj.write_hdf5(mode = 'w')
   traj.create_xml()

   return model, well_interp, datum, traj


@pytest.fixture
def example_model_with_logs(example_model_with_well):

   model, well_interp, datum, traj = example_model_with_well

   frame = WellboreFrame(
      model,
      trajectory = traj,
      represented_interp = well_interp,
      mds = [1, 2, 3, 4],
   )
   frame.write_hdf5()
   frame.create_xml(title = 'Log run A')

   log_collection = frame.logs
   log_collection.add_log("GR", [1, 2, 1, 2], 'gAPI')
   log_collection.add_log("NPHI", [0.1, 0.1, np.NaN, np.NaN], 'v/v')

   return model, well_interp, datum, traj, frame, log_collection


@pytest.fixture
def test_data_path(tmp_path):
   """ Return pathlib.Path pointing to temporary copy of tests/example_data

   Use a fresh temporary directory for each test.
   """
   master_path = (Path(__file__) / '../test_data').resolve()
   data_path = Path(tmp_path) / 'test_data'

   assert master_path.exists()
   assert not data_path.exists()
   copytree(str(master_path), str(data_path))
   return data_path


@pytest.fixture
def example_model_with_properties(tmp_path):
   """Model with a grid (5x5x3) and properties.
   Properties:
   - Zone (discrete)
   - VPC (discrete)
   - Fault block (discrete)
   - Facies (discrete)
   - NTG (continuous)
   - POR (continuous)
   - SW (continuous)
   """
   model_path = str(tmp_path / 'test_no_rels.epc')
   model = Model(create_basics = True, create_hdf5_ext = True, epc_file = model_path, new_epc = True)
   model.store_epc(model.epc_file)

   grid = grr.RegularGrid(parent_model = model,
                          origin = (0, 0, 0),
                          extent_kji = (3, 5, 5),
                          crs_uuid = rqet.uuid_for_part_root(model.crs_root),
                          set_points_cached = True)
   grid.cache_all_geometry_arrays()
   grid.write_hdf5_from_caches(file = model.h5_file_name(file_must_exist = False), mode = 'w')

   grid.create_xml(ext_uuid = model.h5_uuid(),
                   title = 'grid',
                   write_geometry = True,
                   add_cell_length_properties = False)
   model.store_epc()

   zone = np.ones(shape = (5, 5))
   zone_array = np.array([zone, zone + 1, zone + 2])

   vpc = np.array([[1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2], [1, 1, 1, 2, 2]])
   vpc_array = np.array([vpc, vpc, vpc])

   facies = np.array([[1, 1, 1, 2, 2], [1, 1, 2, 2, 2], [1, 2, 2, 2, 3], [2, 2, 2, 3, 3], [2, 2, 3, 3, 3]])
   facies_array = np.array([facies, facies, facies])

   fb = np.array([[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [2, 2, 2, 2, 2], [2, 2, 2, 2, 2]])
   fb_array = np.array([fb, fb, fb])

   ntg = np.array([[0, 0.5, 0, 0.5, 0], [0.5, 0, 0.5, 0, 0.5], [0, 0.5, 0, 0.5, 0], [0.5, 0, 0.5, 0, 0.5],
                   [0, 0.5, 0, 0.5, 0]])
   ntg_array = np.array([ntg, ntg, ntg])

   por = np.array([[1, 1, 1, 1, 1], [0.5, 0.5, 0.5, 0.5, 0.5], [1, 1, 1, 1, 1], [0.5, 0.5, 0.5, 0.5, 0.5],
                   [1, 1, 1, 1, 1]])
   por_array = np.array([por, por, por])

   sat = np.array([[1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1], [1, 0.5, 1, 0.5, 1],
                   [1, 0.5, 1, 0.5, 1]])
   sat_array = np.array([sat, sat, sat])

   collection = rqp.GridPropertyCollection()
   collection.set_grid(grid)
   for array, name, kind, discrete in zip(
      [zone_array, vpc_array, fb_array, facies_array, ntg_array, por_array, sat_array],
      ['Zone', 'VPC', 'Fault block', 'Facies', 'NTG', 'POR', 'SW'],
      ['discrete', 'discrete', 'discrete', 'discrete', 'net to gross ratio', 'porosity', 'saturation'],
      [True, True, True, True, False, False, False]):
      collection.add_cached_array_to_imported_list(cached_array = array,
                                                   source_info = '',
                                                   keyword = name,
                                                   discrete = discrete,
                                                   uom = None,
                                                   time_index = None,
                                                   null_value = None,
                                                   property_kind = kind,
                                                   facet_type = None,
                                                   facet = None,
                                                   realization = None)
      collection.write_hdf5_for_imported_list()
      collection.create_xml_for_imported_list_and_add_parts_to_model()
   model.store_epc()

   return model



"""time.py: A very thin wrapper around python datetime functionality, to meet resqml standard."""

version = '9th April 2020'

import datetime as dt


def now(use_utc = False):
   """Returns an iso format string representation of the current time, to the nearest second.

      argument:
         use_utc (boolean, default False): if True, the current UTC time is used, otherwise local time

      returns:
         string of form YYYY-MM-DDThh:mm:ssZ representing the current time in iso format

      note:
         this is the format used by the resqml standard for representing date-times
   """

   if use_utc:
      stamp = dt.datetime.utcnow()  # NB: naive use of UTC time
   else:
      stamp = dt.datetime.now()
   return str(stamp.isoformat(sep = 'T', timespec = 'seconds')) + 'Z'



# utility.py

version = '29th April 2021'

# simple utility function definitions:
#    def kji0_from_ijk1(index_ijk1):   # reverse order of indices and subtract 1
#    def ijk1_from_kji0(index_kji0):   # reverse order of indices and add 1
#    def extent_switch_ijk_kji(extent_in):   # reverse order of elements in extent (either ijk to kji or vice versa)
#    def string_ijk1_for_cell_kji0(cell_kji0):
#    def string_ijk1_for_cell_ijk1(cell_ijk1):
#    def string_ijk_for_extent_kji(extent_kji):
#    def string_xyz(xyz):
#    def cell_count_from_extent(extent):
#    def horizon_float(k0, plus_or_minus):

import numpy as np


def kji0_from_ijk1(index_ijk1):  # reverse order of indices and subtract 1
   """Returns index converted from simulator protocol to python protocol."""
   dims = index_ijk1.size
   result = np.zeros(dims, dtype = 'int')
   for d in range(dims):
      result[d] = index_ijk1[dims - d - 1] - 1
   return result


def ijk1_from_kji0(index_kji0):  # reverse order of indices and add 1
   """Returns index converted from python protocol to simulator protocol."""
   dims = index_kji0.size
   result = np.zeros(dims, dtype = 'int')
   for d in range(dims):
      result[d] = index_kji0[dims - d - 1] + 1
   return result


def extent_switch_ijk_kji(extent_in):  # reverse order of elements in extent
   """Returns equivalent grid extent switched either way between simulator and python protocols."""
   dims = extent_in.size
   result = np.zeros(dims, dtype = 'int')
   for d in range(dims):
      result[d] = extent_in[dims - d - 1]
   return result


def cell_count_from_extent(extent):
   """Returns the number of cells in a grid with the given extent"""
   result = 1
   for d in range(len(extent)):  # list, tuple or 1D numpy array
      result *= extent[d]
   return result


def string_ijk1_for_cell_kji0(cell_kji0):
   """Returns a string showing indices for a cell in simulator protocol, from data in python protocol."""
   return '[{:}, {:}, {:}]'.format(cell_kji0[2] + 1, cell_kji0[1] + 1, cell_kji0[0] + 1)


def string_ijk1_for_cell_ijk1(cell_ijk1):
   """Returns a string showing indices for a cell in simulator protocol, from data in simulator protocol."""
   return '[{:}, {:}, {:}]'.format(cell_ijk1[0], cell_ijk1[1], cell_ijk1[2])


def string_ijk_for_extent_kji(extent_kji):
   """Returns a string showing grid extent in simulator protocol, from data in python protocol."""
   return '[{:}, {:}, {:}]'.format(extent_kji[2], extent_kji[1], extent_kji[0])


def string_xyz(xyz):
   """Returns an xyz point as a string like (121234.56, 567890.12, 3456.789)"""

   return '({0:4.2f}, {1:4.2f}, {2:5.3f})'.format(xyz[0], xyz[1], xyz[2])


def horizon_float(k0, plus_or_minus):
   """Returns a floating point representation of k direction face index; eg. 3.5 for face between layers 3 and 4."""
   result = float(k0)
   if plus_or_minus == '+':
      result += 0.5
   elif plus_or_minus == '-':
      result -= 0.5
   else:
      assert False
   return result
